package org.zeroturnaround.javarebel.integration.spring;

import org.zeroturnaround.javarebel.Logger;

import org.zeroturnaround.javarebel.LoggerFactory;

import org.zeroturnaround.javarebel.ClassResourceSource;
import org.zeroturnaround.javarebel.IntegrationFactory;
import org.zeroturnaround.javarebel.Plugin;
import org.zeroturnaround.javarebel.integration.spring.cbp.AbstractBeanFactoryCBP;
import org.zeroturnaround.javarebel.integration.spring.cbp.AnnotationMethodHandlerAdapterCBP;
import org.zeroturnaround.javarebel.integration.spring.cbp.CachedIntrospectionResultsCBP;
import org.zeroturnaround.javarebel.integration.spring.cbp.ClassPathBeanDefinitionScannerCBP;
import org.zeroturnaround.javarebel.integration.spring.cbp.ComponentScanBeanDefinitionParserCBP;
import org.zeroturnaround.javarebel.integration.spring.cbp.DefaultAnnotationHandlerMappingCBP;
import org.zeroturnaround.javarebel.integration.spring.cbp.DefaultListableBeanFactoryCBP;
import org.zeroturnaround.javarebel.integration.spring.cbp.DefaultSingletonBeanRegistryCBP;
import org.zeroturnaround.javarebel.integration.spring.cbp.GlobalMethodSecurityBeanDefinitionParserCBP;
import org.zeroturnaround.javarebel.integration.spring.cbp.HttpSecurityBeanDefinitionParserCBP;
import org.zeroturnaround.javarebel.integration.spring.cbp.InjectionMetadataCacheCBP;
import org.zeroturnaround.javarebel.integration.spring.cbp.XmlBeanDefinitionReaderCBP;

/**
 * JavaRebel Spring plugin that manages Spring metadata reloading.
 * 
 * @author Jevgeni Kabanov
 */
public class SpringPlugin implements Plugin{
  private static final boolean enabled = Boolean.getBoolean("rebel.spring_plugin");
  private static final Logger log = LoggerFactory.getInstance();

  public void preinit() {
    if (!enabled) return;

    log.echo("---------------------------------------------------------------------------------------");
    log.echo("JavaRebel: The JavaRebel Spring Framework Plugin is now enabled.");
    log.echo("JavaRebel: Plugins are contributed by third party and can cause compatibility problems.");
    log.echo("JavaRebel: If you have any troubles set -Drebel.spring_plugin=false to disable it.");
    log.echo("---------------------------------------------------------------------------------------");

    ClassLoader cl = SpringPlugin.class.getClassLoader();

    /*
     * Disables Spring annotation cache for dependency method and field annotations.
     */
    IntegrationFactory.getInstance()
    .addIntegrationProcessor(
        cl,
        new String[] {
            "org.springframework.beans.factory.annotation.AutowiredAnnotationBeanPostProcessor",
            "org.springframework.context.annotation.CommonAnnotationBeanPostProcessor",
            "org.springframework.orm.jpa.support.PersistenceAnnotationBeanPostProcessor"
        },
        new InjectionMetadataCacheCBP());       

    /*
     * Disables Spring introspection cache for JavaBeans Introspector results used in
     * resolving XML dependencies.
     */
    IntegrationFactory.getInstance()
    .addIntegrationProcessor(
        cl,
        "org.springframework.beans.CachedIntrospectionResults", 
        new CachedIntrospectionResultsCBP());            

    /*
     * Registers the SpringBeanReloader for each bean factory 
     * and call checkAndReload() on all classes being configured.
     */
    IntegrationFactory.getInstance()
    .addIntegrationProcessor(
        cl,
        "org.springframework.beans.factory.support.AbstractBeanFactory", 
        new AbstractBeanFactoryCBP());       

    /*
     * Disables flushing singleton cache preserving all singleton instances.
     */
    IntegrationFactory.getInstance()
    .addIntegrationProcessor(
        cl,
        "org.springframework.beans.factory.support.DefaultSingletonBeanRegistry", 
        new DefaultSingletonBeanRegistryCBP());      

    /*
     * Registers each XML bean resource with the SpringBeanReloader.
     */
    IntegrationFactory.getInstance()
    .addIntegrationProcessor(
        cl,
        "org.springframework.beans.factory.xml.XmlBeanDefinitionReader", 
        new XmlBeanDefinitionReaderCBP());     

    /*
     * Disables @RequestMapping type-level handler cache.
     */
    IntegrationFactory.getInstance()
    .addIntegrationProcessor(
        cl,
        "org.springframework.web.servlet.mvc.annotation.DefaultAnnotationHandlerMapping", 
        new DefaultAnnotationHandlerMappingCBP());   


    /*
     * Disables @RequestMapping method-level handler cache.
     */
    IntegrationFactory.getInstance()
    .addIntegrationProcessor(
        cl,
        "org.springframework.web.servlet.mvc.annotation.AnnotationMethodHandlerAdapter", 
        new AnnotationMethodHandlerAdapterCBP());   

    /*
     * Disables compatibility check for annotation-configured components 
     * found in classpath (needed from Spring 2.5.5)
     */
    IntegrationFactory.getInstance()
    .addIntegrationProcessor(
        cl,
        "org.springframework.context.annotation.ClassPathBeanDefinitionScanner", 
        new ClassPathBeanDefinitionScannerCBP());

    /*
     * Removes the bean definition to prevent overriding messages 
     * and duplicate exceptions.
     */
    IntegrationFactory.getInstance()
    .addIntegrationProcessor(
        cl,
        "org.springframework.beans.factory.support.DefaultListableBeanFactory", 
        new DefaultListableBeanFactoryCBP());

    /*
     * Avoids error "Duplicate <global-method-security> detected" during a
     * reload of Spring Security bean definitions.
     */
    IntegrationFactory.getInstance()
    .addIntegrationProcessor(
        cl,
        "org.springframework.security.config.GlobalMethodSecurityBeanDefinitionParser",
        new GlobalMethodSecurityBeanDefinitionParserCBP());

    /*
     * Avoids error "Duplicate <http> element detected" during a reload of
     * Spring Security bean definitions.
     */
    IntegrationFactory.getInstance()
    .addIntegrationProcessor(
        cl,
        "org.springframework.security.config.HttpSecurityBeanDefinitionParser",
        new HttpSecurityBeanDefinitionParserCBP());

    /*
     * Scans the classpath for new components.
     */
    IntegrationFactory.getInstance()
    .addIntegrationProcessor(
        cl,
        "org.springframework.context.annotation.ComponentScanBeanDefinitionParser",
        new ComponentScanBeanDefinitionParserCBP());    


    // At the moment hits classloader caching, reenable when problem solved.

    /*IntegrationFactory.getInstance()
    .addIntegrationProcessor(
        cl,
        "org.springframework.beans.factory.config.PropertyResourceConfigurer",
        new PropertyResourceConfigurerCBP()); */      
  }

  public boolean checkDependencies(ClassLoader cl, ClassResourceSource crs) {    
    boolean result = crs.getClassResource("org.springframework.beans.factory.support.AbstractBeanFactory") != null;

    if (result && !enabled) {
      log.echo("---------------------------------------------------------------------------------------");
      log.echo("JavaRebel: You can enable JavaRebel Spring plugin by setting -Drebel.spring_plugin=true.");
      log.echo("JavaRebel: It supports adding new beans and adding new bean dependencies ");
      log.echo("JavaRebel: using annotations or XML. Singletons will be reconfigured after the change.");
      log.echo("JavaRebel: It also supports adding or changing Spring MVC controllers or handlers.");
      log.echo("---------------------------------------------------------------------------------------");
    }

    return result;
  }
}
